
(* reducer *)
open Reducer

(* lts *)
open Automaton_state
open Automaton_arrow
open Automaton_bundle
open Automaton_block
open Automaton_bisimulation
open Automaton_domination
open Automaton_
open Ast
open Parserdecls

(* stdlib *)
open Printf


(* instantiate modules *)
module AutomatonState = State
module AutomatonArrow = Arrow (AutomatonState)
module AutomatonBundle = Bundle (AutomatonState) (AutomatonArrow) 
module MyAutomaton = Automaton (AutomatonState) (AutomatonArrow) (AutomatonBundle)
module AutomatonDomination = Domination (AutomatonArrow) (AutomatonBundle)
module AutomatonBisimulation = Bisimulation (AutomatonBundle)
module AutomatonBlock = Block (AutomatonState) (AutomatonArrow) (AutomatonBundle) (MyAutomaton) (AutomatonBisimulation)
module AutomatonReducer = Reducer (AutomatonState) (AutomatonArrow) (AutomatonBundle) (MyAutomaton) (AutomatonBisimulation) (AutomatonBlock) (AutomatonDomination)


(* [apply roo id] returns the state in the environment roo that
 * is associated with the name id
 *)
let apply roo id =
  try
    (List.assoc id roo)
  with Not_found -> failwith ("apply: not found " ^ id)


(* [create_states dec] returns list of pair (id, state), where
 * the list is created from the declaration dec
 *)
let create_states dec =
  match dec with
      DeclStateList(l) ->
	(List.map 
	   (function 
		DeclState id ->  
		  (id, AutomatonState.create id)
	      | _ -> failwith "create_states: internal error!!")
	   l)
    | _ -> failwith "create_states: internal error!!"


(* [create_arrows dec roo] returns a list of arrows created from
 * the declaration dec and environment roo
 *)
let create_arrows dec roo = 
  match dec with
    | DeclArrowList(l) -> 
	(List.map 
	   (function 
		DeclArrow(id1,id2,label) ->
		  AutomatonArrow.create (apply roo id1) label (apply roo id2)
	      | _ -> failwith "create_arrows: internal error!!")
	   l)
    | _ -> failwith ""

		      
(* [create_start dec roo] returns a state created from the declaration
 * dec and the environment roo
 *)
let create_start dec roo = 
  match dec with
    | DeclStart(id) -> (apply roo id)
    | _ -> failwith "create_start: internal error!!"


(* [create_automata dec] returns an hdautoma from declaration dec *)
let create_automata = 
  function 
    | DeclAutoma(decl_start, decl_states, decl_arrows) -> 
	let roo = (create_states decl_states) in	  
	let arrows = (create_arrows decl_arrows roo) in
	let start = (create_start decl_start roo) in
	  MyAutomaton.create 
	    start
	    (snd (List.split roo))
	    arrows
    | _ -> failwith "create_automa: internal error!!!"


(* [read_file filename] parse the contents of file filename and returns 
 * the abstract syntax tree
 *)
let read_file filename =
  let lexbuf = Lexing.from_channel (open_in filename)
  in Parser.main Lexer.token lexbuf


let _ =
  
  try
    let filename = Sys.argv.(1) in
    let ast = (read_file filename) in
    let automaton = (create_automata ast) in 

      (* starts reduction *)
      let reduced_automaton = (AutomatonReducer.reduce automaton) in 
	begin
	  MyAutomaton.print reduced_automaton
	end

  with
      Failure s -> 
	(fprintf stderr "failure: %s\n" s); (exit ~-1)
    | Invalid_argument s -> 
	(fprintf stderr "usage: lts-reducer filename\n\n"); (exit ~-1)
    | Lexer.Eof ->
	(fprintf stderr "error: unexpected end of file\n"); (exit ~-1)
    | Parsing.Parse_error ->
	(fprintf stderr "error: syntax error at line %d\n" !line_number); (exit ~-1)



