
open Ast
open Printf
open Parserdecls

(* [apply roo id] returns the state in the environment roo that
 * is associated with the name id
 *)
let apply roo id =
  try
    (List.assoc id roo)
  with Not_found -> failwith ("apply: not found " ^ id)

(* [read_file filename] parse the contents of file filename and returns 
 * the abstract syntax tree
 *)
let read_file filename =
  let lexbuf = Lexing.from_channel (open_in filename)
  in Parser.main Lexer.token lexbuf

let print_header f =
  let x = fprintf f in
    x "nets 1\n";
    x "hook \"main\">0\n";
    x "net 0\n";
    x "hook \"automaton\"\n";
    x "hook \"hd-automaton\"\n";
    x "logic \"initial\">0\n\n"

let to_name n = "#" ^ (string_of_int (n-1))
let to_state_list = function DeclStateList l -> l
let to_arrow_list = function DeclArrowList l -> l
let get_state_id = function DeclState( id, _, _ ) -> id

let outdeg id arrows =
  List.length (List.filter (function DeclArrow(source,_,_) -> source = id) arrows)

let outedges id arrows =
  (List.filter (function DeclArrow(source,_,_) -> source = id) arrows)

let rec my_concat sep =
  function
      [] -> ""
    | e::[] -> e
    | e::el -> e ^ sep ^ (my_concat sep el)

let to_label (lab,pi,sigma) =
  match lab with
      "tau" -> "tau"
    | "out" -> let (a,b) = List.nth pi 0, List.nth pi 1 in (to_name a) ^ "!" ^ (to_name b)
    | "in" -> let (a,b) = List.nth pi 0, List.nth pi 1 in (to_name a) ^ "?" ^ (to_name b)
    | "bout" -> 
	let a = List.nth pi 0 in 
	let new_name = Utils.index 1 sigma 0 in
	  (to_name a) ^ "!" ^ (to_name new_name)
    | "bin" -> 
	let a = List.nth pi 0 in 
	let new_name = Utils.index 1 sigma 0 in
	  (to_name a) ^ "?(" ^ (to_name new_name) ^ ")"

let to_hook sigma =
  let new_name = Utils.index 1 sigma 0 in
  let img = List.map (fun x -> (if x = 0 then new_name else x), Utils.index 0 sigma x) sigma in
  let img = List.filter (fun (x,y) -> x!=y) img in 
    my_concat ", " (List.map (fun (x,y) -> (to_name x) ^ "/" ^ (to_name y)) img)

let print_edge f env idx arrow =
  match arrow with
      DeclArrow( source, target, label ) ->
	match label with (lab,pi,sigma) ->
	  let id = (env target) in	  
	    fprintf f " edge %d" idx;
	    fprintf f " behav \"%s\"" (to_label label);
	    fprintf f " hook \"{%s}\"" (to_hook sigma);
	    fprintf f " -> %d\n" id

let print_edges f env arrows =
  let idx = ref 0 in
  let rec p = function
      [] -> ()
    | e::el -> 
	(print_edge f env !idx e);
	idx := !idx + 1;
	(p el) in
    p arrows

let print_node f env arrows node =
  let unquote id = id in
  let print_name_list f names = 
    begin
      fprintf f "\"{%s}\""
	(my_concat "," (List.map to_name names))
    end in

  match node with DeclState(id,names,_) ->
    let n = env id in
      fprintf f "vertex %d" n;
      fprintf f " struct %s" (unquote id);
      fprintf f " behav ";
      print_name_list f names;
      fprintf f " edges %d" (outdeg id arrows);                   (* << outdeg(n); *)
      fprintf f "\n";
      print_edges f env (outedges (get_state_id node) arrows);
      fprintf f "\n"

let export f ast =
  print_header f;
  match ast with
    | DeclAutoma( start_decl, states_decl, arrows_decl ) ->
	let env = Utils.index (-1) (List.map get_state_id (to_state_list states_decl)) in
	let arrows = to_arrow_list arrows_decl in
	(match states_decl with
	   | DeclStateList l ->
	       fprintf f "vertice %d\n\n" (List.length l);
	       List.map (print_node f env arrows) (to_state_list states_decl)

	   | _ -> failwith "*internal error*")
    | _ -> failwith "*internal error*"
	


let _ =
  
  try
    let filename = Sys.argv.(1) in
    let ast = (read_file filename) in
      export stdout ast
  with
      Failure s -> 
	(fprintf stderr "failure: %s\n" s); (exit ~-1)
    | Invalid_argument s -> 
	(fprintf stderr "usage: hd-reducer filename\n\n"); (exit ~-1)
    | Lexer.Eof ->
	(fprintf stderr "error: unexpected end of file\n"); (exit ~-1)
    | Parsing.Parse_error ->
	(fprintf stderr "error: syntax error at line %d\n" !line_number); (exit ~-1)



