(* $Id$ *)

(* Pxp support *)
open Pxp_types;;
open Pxp_document;;
open Pxp_yacc;;

(* open the mihda world *)
open Utils;;
open World;;


let valueOf = 
  function
    | Value n -> n
    | _ -> failwith "value not valid"

let valueOfAttribute attr n = 
  valueOf (n # attribute attr)

let parseStartStates node =
  match node # node_type with
    | T_element "startStates" -> 
	(List.map
	   (valueOfAttribute "stateId")
	   (find_all_elements "start" node))
    | _ -> failwith "parseStartStates: unexpected tag"


let parseName node =
  match node # node_type with
    | T_element "hdsName" -> int_of_string (valueOfAttribute "denotation" node)
    | T_element "newHdsName" -> 0
    | _ -> assert false

let parseMapping node =
  match node # node_type with
    | T_element "mapping" ->
	let (fromNode, toNode) = (find_element "from" node), (find_element "to" node) in 
	  parseName (fromNode # nth_node 0), parseName (toNode # nth_node 0)
    | _ -> assert false
    
let parseWitness node =
  match node # node_type with
    | T_element "witness" ->
        let els = List.map parseMapping (find_all_elements "mapping" node) in
	let ordered_els = List.sort Pervasives.compare els in
	  snd (List.split ordered_els)
    | _ -> assert false

let parsePermutationGroup node =
  match node # node_type with
    | T_element "permutationGroup" ->
	(List.map
	   parseWitness
	   (find_all_elements "witness" node))
    | _ -> assert false

let parseState node =
  let label = ref ""
  and names = ref 0
  and permutationGroup = ref [ [] ]
  in

    match node # node_type with 
      | T_element "state" ->
	  let id = valueOf (node # attribute "id") 
	  and parseInnerTag n =
	    match n # node_type with 
	      | T_element "label" ->            ()
	      | T_element "names" ->            names := int_of_string (valueOfAttribute "nrOf" n)
	      | T_element "permutationGroup" -> permutationGroup := parsePermutationGroup n
	      | T_element _ | T_data ->         ()
	      | _ ->                            assert false
	  in 
	    begin
	      List.iter parseInnerTag (node # sub_nodes);
	      HDState.create id (range 1 (!names+1)) !permutationGroup
	    end
      | _ -> assert false

let parseStates node =
  match node # node_type with
    | T_element "states" -> 
	(List.map
	   parseState
	   (find_all_elements "state" node))
    | _ -> failwith "parseStates: unexpected tag"  

let parseFromTargetNames node =
  match node # node_type with
    | T_element "fromTargetNames" ->
        let els = List.map parseMapping (find_all_elements "mapping" node) in
	let ordered_els = List.sort Pervasives.compare els in
	  snd (List.split ordered_els)
    | _ -> assert false


let positionOfLabelNameSymbol = 
  function
    | "freeInputSubject" -> 0
    | "freeInputObject" -> 1
    | "boundInputSubject" -> 0
    | "freeOutputSubject" -> 0
    | "freeOutputObject" -> 1
    | "boundOutputSubject" -> 0
    | _ -> failwith "positionOfLabelNameSymbol: invalid label"

    
let parseFromLabelName node =
  match node # node_type with
    | T_element "hdsTransitionLabelNameSymbol" -> positionOfLabelNameSymbol (valueOfAttribute "denotation" node)
    | _ -> assert false

let parseMapping2 node =
  match node # node_type with
    | T_element "mapping" ->
	let (fromNode, toNode) = (find_element "from" node), (find_element "to" node) in 
	  parseFromLabelName(fromNode # nth_node 0), parseName (toNode # nth_node 0)
    | _ -> assert false
    
let parseFromNameLabelSymbols node =
  match node # node_type with
    | T_element "fromLabelNameSymbols" -> 
        let els = List.map parseMapping2 (find_all_elements "mapping" node) in
	let ordered_els = List.sort Pervasives.compare els in
	  snd (List.split ordered_els)
    | _ -> assert false

    
let convertPiLabel =
  function
    | "freeInput" -> "in"
    | "boundInput" -> "bin"
    | "freeOutput" -> "out"
    | "boundOutput" -> "bout"
    | "tau" -> "tau"
    | _ -> failwith "convertPiLabel: invalid label"
      
let convertFromPiLabel =
  function
    | "in" -> "freeInput" 
    | "bin" -> "boundInput" 
    | "out" -> "freeOutput" 
    | "bout" -> "boundOutput" 
    | "tau" -> "tau"
    | s -> failwith( "convertFromPiLabel: invalid label " ^ s )
      
let parseTransition sourceId env node =

  let label = ref ""
  and mappingsToSourceNames = ref []
  and fromLabelNameSymbols = ref []
  and fromTargetNames = ref []
  in

    match node # node_type with
      | T_element "transition" -> 
	  let parseInnerTag n =
	    match n # node_type with
	      | T_element "label" ->                     
		  label := convertPiLabel (valueOfAttribute "kind" (find_element "hdsOverPiAction" n))
	      | T_element "mappingsToSourceNames" ->      
		  fromLabelNameSymbols := [];
		  fromLabelNameSymbols := parseFromNameLabelSymbols (find_element "fromLabelNameSymbols" n);
		  fromTargetNames := parseFromTargetNames (find_element "fromTargetNames" n)
	      | T_data -> () (* ignore *)
	      | T_element _ | _ ->               
		  assert false  
	  in 
	    begin
	      List.iter parseInnerTag (node # sub_nodes);
	      HDArrow.create (env sourceId) (!label, !fromLabelNameSymbols, !fromTargetNames) (env (valueOfAttribute "targetId" node))
	    end
      | _ -> assert false

let parseBundle env node =
  match node # node_type with
    | T_element "bundle" -> 
	(List.map
	   (parseTransition (valueOfAttribute "sourceId" node) env)
	   (find_all_elements "transition" node))
    | _ -> assert false

let parseTransitions env node =
  match node # node_type with
    | T_element "transitions" -> 
	List.flatten (List.map
			(parseBundle env)
			(find_all_elements "bundle" node))
    | _ -> failwith "parseTransitions: unexpected tag"  

let stateById states id =
  (List.find
     (fun st -> (HDState.id st) = id)
     states)

let parseTransitionSystem node =

  let startStates = ref [] 
  and states = ref [] 
  and transitions = ref [] 
  in
    
    match node # node_type with 
      | T_element "transitionSystem" -> 
	  let parseInnerTag n = 
	    match n # node_type with
	      | T_element "startStates" ->     startStates := parseStartStates n
	      | T_element "states" ->          states := parseStates n
	      | T_element "transitions" ->     transitions := parseTransitions (stateById !states) n
	      | T_element "hdsOrdering" ->     () (* ignore *)
	      | T_data ->                      () (* ignore *)
	      | T_element e ->                 failwith ("unexpected tag " ^ e)
	      | _ ->                           failwith ("!!!!")
	  in 
	    begin
	      List.iter parseInnerTag (node # sub_nodes);
	      HD.create 
		(stateById !states (List.nth !startStates 0) )
		!states
		!transitions
	    end
      | _ -> failwith "parseTransitionSystem: unexpected tag"


class warner =
object
  method warn w =
    Printf.fprintf stdout "warning: %s\n" w
end

let parse_hdxml file =
  try
    let config = { default_config with warner = new warner } in
    let document = parse_wfcontent_entity config file default_spec in
    let root = document # root in
      parseTransitionSystem root
  with
      x ->
	prerr_endline (string_of_exn x);
	exit 1
;;

let print_mapping_xml file l =
   let p = Printf.fprintf in
     p file "  <witness>\n";
     (List.iter2 
       (fun i v ->
          p file "    <mapping>\n"; 
	  p file "      <from><hdsName denotation=\"%d\" /></from>\n" i;
	  p file "      <to><hdsName denotation=\"%d\" /></to>\n" v;
	  p file "    </mapping>\n") (range 1 ((List.length l)+1)) l);
     p file "  </witness>\n"
   
let print_state_xml file state =
   let p = Printf.fprintf in
     p file "<state id=\"%s\">\n" (HDState.id state);
     p file "  <names nrOf=\"%d\" />\n" (List.length (HDState.names state));
     p file "  <permutationGroup>\n";
     List.iter (print_mapping_xml file) (HDState.group state);
     p file "  </permutationGroup>\n";
     p file "</state>\n"
     
let print_from_target_names_xml file quad = 
   let l = HDBundle.sigma quad in
   let p = Printf.fprintf in
     p file "      <fromTargetNames>\n";
     (List.iter2 
       (fun i v ->
          p file "        <mapping>\n"; 
	  p file "          <from><hdsName denotation=\"%d\" /></from>\n" i;
	  (match v with
	    | 0 -> p file "          <to><newHdsName /></to>\n"
	    | _ -> p file "          <to><hdsName denotation=\"%d\" /></to>\n" v);
	  p file "        </mapping>\n") (range 1 ((List.length l)+1)) l);
     p file "      </fromTargetNames>\n"
     
     
let print_label_symbol file kind v =
   let p = Printf.fprintf in
     p file "        <mapping>\n"; 
     p file "          <from><hdsTransitionLabelNameSymbol denotation=\"%s\" /></from>\n" kind;
     p file "          <to><hdsName denotation=\"%d\" /></to>\n" v;
     p file "        </mapping>\n"
  
let print_from_label_name_symbols file quad =
   let pi = HDBundle.pi quad in
   let p = Printf.fprintf in
     p file "      <fromLabelNameSymbols>\n";
     (match (HDBundle.label quad) with
        | "in" -> 
	  print_label_symbol file "freeInputSubject" (HDBundle.subj quad);
	  print_label_symbol file "freeInputObject" (HDBundle.obj quad)
	| "bin" -> 
	  print_label_symbol file "boundInputSubject" (HDBundle.subj quad);
	| "out" ->
	  print_label_symbol file "freeOutputSubject" (HDBundle.subj quad);
	  print_label_symbol file "freeOutputObject" (HDBundle.obj quad)
	| "bout" ->
	  print_label_symbol file "boundOutputSubject" (HDBundle.subj quad);
	| "tau" -> ()
	| _ -> failwith "print_from_label_name_symbols: invalid label");
     p file "      </fromLabelNameSymbols>\n"
     
let print_quad_xml file quad =
   let p = Printf.fprintf in
     p file "  <transition targetId=\"%s\">\n" (HDState.id (HDBundle.target quad));
     p file "    <label>\n";
     p file "      <hdsOverPiAction kind=\"%s\" />\n" (convertFromPiLabel (HDBundle.label quad));
     p file "    </label>\n";
     p file "    <mappingsToSourceNames>\n";
     print_from_label_name_symbols file quad;
     print_from_target_names_xml file quad;
     p file "    </mappingsToSourceNames>\n";
     p file "  </transition>\n"
     
let print_bundle_xml file state bundle =
   let p = Printf.fprintf in
     p file "<bundle sourceId=\"%s\">\n" (HDState.id state);
     List.iter (print_quad_xml file) bundle;
     p file "</bundle>\n"
     
let print_hdxml file hda =
   let p = Printf.fprintf in
   let (start, states, arrows) = HD.start hda, HD.states hda, HD.arrows hda in
   let bundles = List.map (HD.bundle hda) states in
      p file "<transitionSystem>\n";
      p file "<startStates>\n";
      p file "  <start stateId=\"%s\" />\n" (HDState.id start);
      p file "</startStates>\n";
      p file "<states>\n";
      List.iter (print_state_xml file) states;
      p file "</states>\n\n";
      p file "<transitions>\n";
      List.iter2 (print_bundle_xml file) states bundles;
      p file "</transitions>\n";
      p file "<hdsOrdering spec=\"lexicographic\" />\n";
      p file "</transitionSystem>\n"
      
      
      
      
      
