open Hd_state
open Hd_arrow
open Hd_bundle
open Hd_automaton
open Hd_bisimulation
open Utils
open Printf

module type BlockSig =
sig
  type automatontype
  and statetype
  and arrowtype
  and bundletype
  and buckettype
  and resulttype
  and t

  val from_states : statetype list -> t
  val id : t -> string
  val states : t -> statetype list
  val norm : t -> bundletype
  val names : t -> int list
  val group : t -> int list list
  val sigmas : t -> statetype -> (int * int) list list
  val inv_thetas : t -> statetype -> (int * int) list
  val create :
    string ->
      statetype list ->
        bundletype ->
          int list ->
            int list list ->
              (statetype -> (int * int) list list) ->
		(statetype -> (int * int) list) -> t
  val compare : t -> t -> int
  val cardinal : t -> int
  val to_state : int -> t -> statetype
  val mem : statetype -> t -> bool
  val new_sigma :
    (statetype -> t) -> statetype -> (int * int) list list
  val split : bundletype -> (statetype -> resulttype option) -> t -> buckettype * t
  val close_block : (statetype -> t) -> buckettype -> t
  val next :
    (statetype -> t) ->
      (t ->statetype) -> bundletype -> bundletype
end


module Block =
  functor(State : StateSig) ->
  functor(Arrow : ArrowSig with type statetype = State.t) ->
  functor(Bundle : BundleSig with type statetype = Arrow.statetype and type arrowtype=Arrow.t) ->
  functor(Automaton : AutomatonSig with type bundletype = Bundle.quadtype list and type statetype = State.t) ->
  functor(Bisimulation : BisimulationSig with type bundletype = Bundle.quadtype list) ->
struct
    type automatontype = Automaton.t
    and statetype = State.t
    and arrowtype = Automaton.arrowtype
    and bundletype = Bundle.quadtype list
    and buckettype =
	Bucket of
	  statetype list *                (* list of the states in the bucket *)
	  bundletype     *                (* the normalized bundle of the bucket *)
	  int list       *                (* active names of the bucket *)
	  (statetype -> (int * int) list) (* thetas of the bucket *)
    and resulttype = Bisimulation.resulttype
    and t =
	Block of 
	  string *
	  statetype list *
	  bundletype *
	  int list *
	  int list list *
	  (statetype -> (int * int) list list) *
	  (statetype -> (int * int) list)
	    
    let from_states states =
      Block("", states, (Bundle.from_arrow_list []), [], [], (fun x -> [ [] ]), (fun x -> []))

    let id = function Block(name, states, norm, names, group, sigmas, inv_thetas) -> name
    let states = function Block(name, states, norm, names, group, sigmas, inv_thetas) ->states
    let norm = function Block(name, states, norm, names, group, sigmas, inv_thetas) -> norm
    let names = function Block(name, states, norm, names, group, sigmas, inv_thetas) -> names
    let group = function Block(name, states, norm, names, group, sigmas, inv_thetas) -> group
    let sigmas = function Block(name, states, norm, names, group, sigmas, inv_thetas) -> sigmas
    let inv_thetas = function Block(name, states, norm, names, group, sigmas, inv_thetas) -> inv_thetas

    let create name states norm names group sigmas inv_thetas =
      Block(name, states, norm, names, group, sigmas, inv_thetas)

    let cardinal block = List.length (states block)

    let to_state n block =
      State.create
	("b" ^ (string_of_int n))
	(names block)
	(group block)
       
    let mem state block =
      List.mem state (states block)
	    
    let new_sigma new_inv_thetas new_group env =
      fun q ->
	try
	  let q_block = env q in
	  let q_group = new_group in (* (group q_block) in *)
	  let q_inv_thetas = new_inv_thetas q in
	    (List.map 
	       (fun perm -> 
		  let f = compose (perm_to_fun perm) (to_fun q_inv_thetas) in
		    List.map (fun x->(x, (f x))) (range 1 (1+(List.length perm))))
	       q_group)
	with Not_found -> failwith "new_sigma: ..."


    let split minimal pred block =
      let map = List.map (fun q -> q, (pred q)) (states block) in
      let (q_bis, q_non_bis) = List.partition (fun (q,th) -> th != None) map in			       (* a + s * round(a/s) + O( round(a/s) ^ 2 ) + O( 2 * round(a/s) + 4 * p * round(a/s) ^ 3 ) *)
      let new_states =  (fst (List.split q_bis)) in						       (* a + s * round(a/s) + O( round(a/s) ^ 2 ) + O( 2 * round(a/s) + 4 * p * round(a/s) ^ 3 ) *)
      let old_states =  (fst (List.split q_non_bis)) in						       (* a + s * round(a/s) + O( round(a/s) ^ 2 ) + O( 2 * round(a/s) + 4 * p * round(a/s) ^ 3 ) *)
      let new_inv_thetas = fun q ->
	try 
 	  invert ((function Some x->x) (List.assoc q map) )
	with Not_found -> failwith "new_inv_thetas: exception" in
	(Bucket(new_states, minimal, (Bundle.active_names minimal), new_inv_thetas),
	 (create
	    (id block)
	    old_states
	    (norm block)
	    (names block)
	    (group block)
	    (sigmas block)
	    (inv_thetas block)))

    let close_block env = function Bucket(states, norm, an, inv_thetas) ->
      let new_group = Bundle.compute_group an norm (permutations (List.length an)) in
      let new_sigma = new_sigma inv_thetas new_group env in
	create
	  ""
	  states
	  norm
	  an
	  new_group
	  new_sigma
	  inv_thetas


    let next env h_n bundle =
      let renaming = compose env h_n in
	List.sort
	  compare
	  (unique
	     (List.flatten
		(List.map 
		   (fun qd -> 
		      let target = (Bundle.target qd) in
		      let block = env target in
		      let sigmas = sigmas block in
			List.map
			  (fun sigma' -> (Bundle.apply_theta (Bundle.apply_dest renaming qd) sigma'))
			  (sigmas target))
		bundle )))


    let compare block1 block2 = 
      let b1 = List.sort Bundle.compare (norm block1) in
      let b2 = List.sort Bundle.compare (norm block2) in
      let an1 = List.length (names block1) in
      let an2 = List.length (names block2) in
	compare (an1,b1) (an2,b2)

	

  end
