
open Hd_state
open Hd_arrow
open Utils
open Printf

module type BundleSig = 
sig
  type arrowtype
  and statetype
  and labeltype
  and quadtype
  and t = quadtype list 
  val names : quadtype -> int list
  val target : quadtype -> statetype
  val label : quadtype -> string
  val pi : quadtype -> int list
  val sigma : quadtype -> int list
  val labels : quadtype list -> string list
  val obj : quadtype -> int
  val subj : quadtype -> int
  val from_arrow_list : arrowtype list -> quadtype list
  val to_arrow_list : statetype -> quadtype list -> arrowtype list
  val apply_dest :
    (statetype -> statetype) -> quadtype -> quadtype
  val apply_theta : quadtype -> (int * int) list -> quadtype
  val active_names : quadtype list -> int list
  val remove_in : int list -> quadtype -> bool
  val apply_perm  : int list -> quadtype -> int list -> quadtype
  val apply_perm2 : int list -> quadtype -> int list -> quadtype
  val create_qd : labeltype -> statetype -> quadtype
  val apply_perm_to_bundle2 :
    int list -> quadtype list -> int list -> quadtype list
  val normalize :
    (quadtype list -> quadtype list) -> quadtype list -> quadtype list
  val minimize :
    (quadtype list -> quadtype list) -> quadtype list -> quadtype list
  val rename : quadtype list -> quadtype list
  val diff : 'a list -> 'a list -> 'a list
  val compare : 'a -> 'a -> int
  val print_qd : quadtype -> unit
  val print : quadtype list -> unit
  val compute_group :
    int list -> quadtype list -> int list list -> int list list
end



module Bundle =
  functor(State : StateSig) ->
  functor(Arrow : ArrowSig with type statetype = State.t) ->
struct
  type arrowtype = Arrow.t
  and statetype = Arrow.statetype
  and labeltype = Arrow.labeltype
  and quadtype = Qd of labeltype * statetype
  and t = quadtype list

  let names = function Qd((lab,pi,sigma),target) -> (pi @ sigma)

  let target = function Qd(lab, target) -> target

  let label  = function Qd((lab,pi,sigma), target) -> lab

  let pi  = function Qd((lab,pi,sigma), target) -> pi

  let sigma  = function Qd((lab,pi,sigma), target) -> sigma

  let labels bundle = Utils.unique (List.map label bundle)

  let obj  = function Qd((lab,pi,sigma),target) ->
    match lab with
      | "in" | "out" | "bin" | "bout" -> (try List.nth pi 1 with _ -> failwith "obj: no object")
      | _ -> failwith "obj: no object available"

  let subj = function Qd((lab,pi,sigma),target) ->
    match lab with
      | "in" | "out" | "bin" | "bout" -> (try List.nth pi 0 with _ -> failwith "subj: no subject")
      | _ -> failwith "subj: no subject available"

  let rec from_arrow_list = function
    | [] -> []
    | ar::arws -> Qd((Arrow.label ar), (Arrow.target ar)) :: (from_arrow_list arws)
				 
  let rec to_arrow_list s = function
    | [] -> []
    | Qd(l,t)::qds -> (Arrow.create s l t) :: (to_arrow_list s qds)

  let apply_dest f = function
    | Qd(lab, target) -> Qd(lab, (f target))


  let print_qd = function Qd((lab,pi,sigma), t) ->
    Printf.fprintf stdout "  label = %s" lab;
    Printf.fprintf stdout "  pi = "; print_int_list stdout pi;
    Printf.fprintf stdout "  sigma = "; print_int_list stdout sigma;
    Printf.fprintf stdout "  target = "; State.print t

  let apply_theta qd sigma' =
    match qd with
	Qd((lab, pi, sigma), target) ->
	  let f i = List.nth sigma (((to_fun sigma') i) - 1) in
	  let dom_sigma = range 1 (1+List.length (dom sigma')) in
	  let new_sigma = List.map f dom_sigma in 
	    Qd((lab, pi, new_sigma), target)
			       
  let active_names bundle =
    Utils.unique (Utils.list_rem 0 (List.flatten (List.map names bundle)))
	
	     
  let remove_in an =
    function Qd((lab,pi,sigma),target) as qd ->
      (lab = "in") && (not (List.mem (obj qd) an))
			  

  let print = List.iter print_qd


  let apply_perm names_ qd perm =
    match qd with Qd((lab, pi, sigma),target) -> 
      let my_index list el = if (el == 0) then 0 else Utils.index 0 list el in
	Qd((let names = (List.sort compare names_) in
	    let fun_norm = perm_to_fun names in 
 	    let new_pi = pi in
	    let sigma_norm = List.map (my_index names) sigma in
	    let sigma_perm = Utils.compose_perm sigma_norm perm in
	    let new_sigma = List.map fun_norm sigma_perm in
	      (lab, new_pi, new_sigma)),
	   (target))



  (* [apply_perm2 names_ ar perm] considers names_ as the identity permutation of names_
   * and returns the application of the permutation perm to the arrow ar 
   *)
  let apply_perm2 names_ qd perm =
    match qd with Qd((lab, pi, sigma),target) -> 
      let my_index list el = if (el == 0) then 0 else Utils.index 0 list el in
	Qd((let names = (List.sort compare names_) in
	    let fun_norm = perm_to_fun names in 
	    let pi_norm = List.map (my_index names) pi in
	    let pi_perm = Utils.compose_perm pi_norm perm in
 	    let new_pi = List.map fun_norm pi_perm in
	    let sigma_norm = List.map (my_index names) sigma in
	    let sigma_perm = Utils.compose_perm sigma_norm perm in
	    let new_sigma = List.map fun_norm sigma_perm in
	      (lab, new_pi, new_sigma)),
	   (target))

  let create_qd lab target = Qd(lab, target)


  let apply_perm_to_bundle2 names bl perm =
    List.map (fun x -> apply_perm2 names x perm) bl
      

	
  let rename bundle =
    let an = 0 :: ( List.sort compare (active_names bundle) ) in
    let rec find_index l j x =
      match l with
	| [] -> failwith "find_index: empty list!"
	| el::els -> if (el = x) then j else ( find_index els (j + 1) x ) in

    let rename_quad = function Qd((pilabel, pi, sigma), target) ->
      let new_pi = List.map (find_index an 0) pi in
      let new_sigma = List.map (find_index an 0) sigma in
	  Qd((match pilabel with
	     | "tau" | "in" | "bin" | "out" | "bout" -> 
		 (pilabel, new_pi, new_sigma)
	     | _ -> failwith "rename_arrow: unknown label"),
	     target) in
      List.map rename_quad bundle


  let normalize red bundle =
    let w_bundle = red bundle in
    let an = active_names w_bundle in
      rename (list_diff bundle (List.filter (remove_in an) bundle))
	
  let minimize red bundle =
    let w_bundle = red bundle in
    let an = (active_names w_bundle) in
    let b = rename (list_diff bundle (List.filter (remove_in an) bundle)) in
    let names = (active_names b) in
    let perms = permutations (List.length names) in
    let bundles = List.sort compare (List.map (apply_perm_to_bundle2 names b) perms) in
      try
	List.hd bundles
      with _ -> failwith "norm: the list of bundles is empty!"
	  
  let diff = Utils.list_diff

  let compare = compare

  let rec compute_group an bundle =
    function
      | [] -> []
      | rho :: rhoes -> 
	  let unchanged = 
	    (List.for_all 
	       (fun ar ->
		  (List.mem (apply_perm2 an ar rho) bundle))
	       bundle) in
	    if unchanged
	    then rho::(compute_group an bundle rhoes)
	    else (compute_group an bundle rhoes)
	      
end

