
(* reducer *)
open Reducer

(* hd automaton *)
open Hd_state
open Hd_arrow
open Hd_bundle
open Hd_block
open Hd_bisimulation
open Hd_automaton
open Hd_domination

open Ast

open Printf

open Parserdecls

module HDState = State
module HDArrow = Arrow (HDState)
module HDBundle = Bundle (HDState) (HDArrow) 
module MyHD = Automaton (HDState) (HDArrow) (HDBundle)
module HDBisimulation = Bisimulation (HDBundle)
module HDBlock = Block (HDState) (HDArrow) (HDBundle) (MyHD) (HDBisimulation)
module HDDomination = Domination (HDBundle)
module HDReducer = Reducer (HDState) (HDArrow) (HDBundle) (MyHD) (HDBisimulation) (HDBlock) (HDDomination)




(* [apply roo id] returns the state in the environment roo that
 * is associated with the name id
 *)
let apply roo id =
  try
    (List.assoc id roo)
  with Not_found -> failwith ("apply: not found " ^ id)

let convertName id =
  if String.contains id '"' 
  then String.sub id 1 ((String.length id) - 2)
  else id

(* [create_states dec] returns list of pair (id, state), where
 * the list is created from the declaration dec
 *)
let create_states dec =
  match dec with
      DeclStateList(l) ->
	(List.map 
	   (function 
		DeclState (id, names, group) ->  
		  ((convertName id), HDState.create (convertName id) names group)
	      | _ -> failwith "create_states: internal error!!")
	   l)
    | _ -> failwith "create_states: internal error!!"


(* [create_arrows dec roo] returns a list of arrows created from
 * the declaration dec and environment roo
 *)
let create_arrows dec roo = 
  match dec with
    | DeclArrowList(l) -> 
	(List.map 
	   (function 
		DeclArrow(id1,id2,label) ->
		  HDArrow.create (apply roo (convertName id1)) label (apply roo (convertName id2))
	      | _ -> failwith "create_arrows: internal error!!")
	   l)
    | _ -> failwith ""

		      
(* [create_start dec roo] returns a state created from the declaration
 * dec and the environment roo
 *)
let create_start dec roo = 
  match dec with
    | DeclStart(id) -> (apply roo (convertName id))
    | _ -> failwith "create_start: internal error!!"


(* [create_automata dec] returns an hdautoma from declaration dec *)
let create_automata = 
  function 
    | DeclAutoma(decl_start, decl_states, decl_arrows) -> 
	let roo = (create_states decl_states) in	  
	let arrows = (create_arrows decl_arrows roo) in
	let start = (create_start decl_start roo) in
	  MyHD.create 
	    start
	    (snd (List.split roo))
	    arrows
    | _ -> failwith "create_automa: internal error!!!"


(* [read_file filename] parse the contents of file filename and returns 
 * the abstract syntax tree
 *)
let read_file filename =
  let lexbuf = Lexing.from_channel (open_in filename)
  in Parser.main Lexer.token lexbuf

let read_stdin () =
  let lexbuf = Lexing.from_channel stdin
  in Parser.main Lexer.token lexbuf

let _ =
  
  try
(*    let filename = Sys.argv.(1) in *)
(*    let ast = (read_file filename) in *)
    let ast = read_stdin() in 
    let automaton = (create_automata ast) in 

      (* starts reduction *)
      let reduced_automaton = (HDReducer.reduce automaton) in 
	begin
	  MyHD.print reduced_automaton
	end

  with
      Failure s -> 
	(fprintf stderr "failure: %s\n" s); (exit ~-1)
    | Invalid_argument s -> 
	(fprintf stderr "usage: hd-reducer filename\n\n"); (exit ~-1)
    | Lexer.Eof ->
	(fprintf stderr "error: unexpected end of file\n"); (exit ~-1)
    | Parsing.Parse_error ->
	(fprintf stderr "error: syntax error at line %d\n" !line_number); (exit ~-1)



