(*s The module [Block] is the signature for blocks.
  * A block is the data structure which contains
  * the states that are considered equivalent at a given
  * iteration.
  * The main operation on a block is the split operation
  * that divide a block into buckets, i.e. quasi-blocks
  * that have some components that should be uniformely computed
  * at the end of the splitting phase.
  *
  * [Reducer] will return a list of blocks as result of each
  * iteration. Such block represents the states of the current
  * approximation of the minimal automaton.
 *)
module type Block =
sig

  (*s is the type of states.
   *)
  type statetype

  (*s is the type of bundles.
   *)
  type bundletype

  (*s is the type of buckets.
   *)
  type buckettype

  (*s is the type used by the splitting operation
    * to separate the states of a block into different
    * equivalence classes.
   *)
  type resulttype

  (*s is the type of a block.
   *)
  type t

  (*s returns the name of [block].
    *
    * @note It is assumed that blocks have unique identifiers.
   *)
  val id : block:t -> string

  (*s returns the list of states in [block].
   *)
  val states : block:t -> statetype list

  (*s returns the cardinality of the set of states in [block].
   *)
  val cardinal : block:t -> int

  (*s returns the normalized bundle of [block].
   *)
  val norm : block:t -> bundletype

  (*s returns [true] if, and only if, [state] is member of
    * the set of states in [block].
   *)
  val mem : state:statetype -> block:t -> bool

  (*s builds a block out of a list of states.
    *
    * @todo rename from_states to init/initialize
   *)
  val from_states : states:(statetype list) -> t

  (*s converts a [block] into a state.
    * Integer [id] is used to uniquely generate the name of [block].
   *)
  val to_state : id:int -> block:t -> statetype

  (*s converts [buck] into a block. The conversion requires
    * an environment [env] that associates to a state its
    * containing block.
   *)
  val close_block: env:(statetype -> t) -> buck:buckettype -> t

  (*s returns the application of [h_n], the [n]-th approximation of
    * the functor (see [FMP02]) to [bundle].
    * As for [to_state] and environment [env] is required in order
    * to substitute the destination states on [bundle] with the
    * block that contains them.
   *)
  val next : env:(statetype -> t) -> h_n:(t -> statetype) -> bundle:bundletype -> bundletype

  (*s separates the states of [block] whose normalized bundle is to [bundle]
    * equivalent, according to [pred].
    * Indeed, predicate [pred] returns [None] if such equivalence does not
    * hold, otherwise it returns [Some(r)], where [r] establishes
    * the correspondence between the two bundles.
    * The result is a pair bucket-block, where the first component
    * is the bucket made of the equivalent states and the second
    * component is the [block] where such states are removed.
   *)
  val split : bundle:bundletype -> pred:(statetype -> resulttype option) -> block:t -> buckettype * t

  (*s compares blocks [blk1] and [blk2].
    * The result is
    *    @il
    *      @i 0, if the blocks are equal,
    *      @i -1, if [blk1] is less than [blk2],
    *      @i 1, otherwise.
    *    @li
    *
    * @arg [blk1] is the first block
    * @arg [blk2] is the second block
    *
    * @note The comparison is meant to be a structural comparison. However, any function that does not equate two conceptually different states may be adopted.
  *)
  val compare: blk1:t -> blk2:t -> int

  (* @todo add print *)
end

