(*s The module [Bundle] defines the interface for bundle types.
  * A bundle contains the information about the observables
  * and future states carried out by the transitions starting
  * from a given state.
  * The module relies on [statetype], [arrowtype] and [quadtype].
  *
  * Usually, a bundle is computed from a list of arrows.
 *)
module type Bundle =
sig
  (*s is the type of the arrow.
    *
    * @see [from_arrow_list], [to_arrow_list] and [Automaton.bundle].
   *)
  type arrowtype

  (*s is the type of the elements in the bundle.
    *
    * @note the bundle's type is a list of [quadtype].
   *)
  type quadtype

  (*s is the type of the states.
   *)
  type statetype

  (*s creates a bundle from the list of transitions  [arrows].
   *)
  val from_arrow_list : arrows:(arrowtype list) -> quadtype list

  (*s returns a list of arrows with source [state] from [bundle].
    * This functions is used by [Reducer] to compute the arrows
    * of the minimal automaton.
    *
    * @see [Reducer], [Block.norm].
   *)
  val to_arrow_list : state:statetype -> bundle:(quadtype list) -> arrowtype list

  (*s returns a normalized bundle from [bundle] and the reduce function [red].
    *
    * @see [Bisimultation.bisimilar]
   *)
  val normalize : red:(quadtype list -> quadtype list) -> bundle:(quadtype list) -> quadtype list

  (*s returns a minimized bundle from [bundle] and the reduce function [red].
    * Function [red] is supposed to eliminate dominated transitions from
    * a given bundle (see Domination)
    * We underline that the minimization is parameterized by the [compare]
    * functions on states, arrows and quadruples.
    *
    * [Reducer] uses [minimize] to compute the representative bundle
    * of a block.
    *
    * @see [Bisimultation.bisimilar] and [Block]
   *)
  val minimize : red:(quadtype list -> quadtype list) -> bundle:(quadtype list) -> quadtype list

  (*s returns the bundle obtained by [bl1] minus all quadruples in [bl2].
   *)
  val diff : bl1:(quadtype list) -> bl2:(quadtype list) -> quadtype list

  (*s compares the two bundles [bl1] and [bl2].
    * The result is
    *    @il
    *     @i 0, if the bundles are equal,
    *     @i -1, if [bl1] is less than [bl2],
    *     @i 1, otherwise.
    *    @li
    *
    * @arg [bl1] is the first block
    * @arg [bl2] is the second block
    *
    * @note The comparison is meant to be a structural comparison. However, any function that does not equate two conceptually different states may be adopted.
   *)
  val compare : bl1:(quadtype list) -> bl2:(quadtype list) -> int

  (*s prints [bundle] on the standard output.
   *)
  val print : bundle:(quadtype list) -> unit

end
