(*i The module [Reducer] contains the main function (i.e. [reduce])
   and the minimization cicle.
   At the end of the computation, the minimized automaton is built.
 *)

(*i*)
open State
open Arrow
open Bundle
open Block
open Automaton
open Bisimulation
open Domination
open Utils

module Reducer =

  (*s [Reducer] depends on all the the modules.
    * Constraits on type dependencies are specified by
    * the following functors.
   *)
  functor (State : State) ->
  functor (Arrow : Arrow with type statetype = State.t ) ->
  functor (Bundle: Bundle with type statetype = State.t and type arrowtype = Arrow.t) ->
  functor (Automaton : Automaton with type statetype = State.t and type arrowtype = Arrow.t and type bundletype = Bundle.quadtype list) ->
  functor (Bisimulation : Bisimulation with type bundletype = Bundle.quadtype list) ->
  functor (Block : Block with type statetype = State.t and type bundletype = Bundle.quadtype list and type resulttype = Bisimulation.resulttype) ->
  functor (Domination : Domination with type quadtype = Bundle.quadtype) ->

struct
  let blocks = ref []

  (*s [blocks_to_states] [hd] [id] [blocks] returns a list of states from
    * [blocks],
    *
    * @note: for now the created states have no names!!!
   *)
  let rec blocks_to_states n =
    function
      | [] -> []
      | blk::blks ->
	  (blk, Block.to_state n blk)
	  :: (blocks_to_states (n+1) blks)

  let states_n blocks = (blocks_to_states 0 blocks)
		 
  let h_n blocks block =  List.assq block (states_n blocks)

  let env blocks s =
    try
      List.find (Block.mem s) blocks
    with Not_found ->
      if blocks <> []
      then failwith "env: state not found"
      else failwith "env: blocks are empty"

  let red bundle =
    let dominated =
      List.filter
	(fun qd -> None <> (Domination.dominated qd bundle))
	bundle in
      list_diff bundle dominated

  (* partitioning [aut] returns a partitioned list of blocks from the
   * automata [aut].
   * NOTE: states in the same block are bisimilar in the given iteration.
   *)
(*i*)

  (* This Section deals with the implementation of the partitioning
     algorithm. In particular, the main part of [Reducer]'s code are
     described.
  *)
    
  let partitioning aut = 
    let start,states, arrows = (Automaton.start aut),
                               (Automaton.states aut),
			       (Automaton.arrows aut) in

      (* Initially, the list of [blocks] is made of a single block
	 that contains all  automaton's states.
      *)
      blocks := [ (Block.from_states states) ] ;

      (* [split blocks block] returns a pair ([bucket],[block']) where
	 [bucket] contains all the states supposed equivalent at the
	 current iteration and [block'] is obtained by removing those
	 states from [block].
      *)
      let split blocks block =
	try

	  (* [minimal] computes the minimal bundle of the first state
	     of [block]. 
	     Note that to compute the [minimal] and the [normalized] bundle
	     we use three auxiliary functions [red], [env] and [h_n].
	     [red] is a filter function, e.g., for a given
	     bundle [b] it returns the bundle obtained by removing from [b]
	     all dominated quadruples (see {\em Domination.dominated}).
	     [env] maps states to blocks; in particular given a state [q]
	     returns the block that approximate [q]
	     [h_n] maps blocks to states; given a block [b] returns
	     the states [q] that represent [b] in the [n]-th approximation.
	   *)
	  let minimal =
	    (Bundle.minimize red
	       (Block.next
		  (env blocks)
		  (h_n blocks)
		  (Automaton.bundle aut (List.hd (Block.states block))))) in

	    (* 
	       At this point, [block] is splitted in the pair (bucket, block').
	       More precisely, the function [Block.split] is invoked with a
	       predicate that, for each state [q], computes its normalized bundle
	       [normal] and returns ([Bisimulation.bisimilar] [minimal] [normal]).
	    *)
	    Some (Block.split
		    minimal
		    (fun q ->
		       let normal =
			 (Bundle.normalize
			    red
			    (Block.next (env blocks)
			       (h_n blocks)
			       (Automaton.bundle aut q))) in
			 Bisimulation.bisimilar minimal normal)
		    block)
	with Failure e -> None in

	
      (* [split_iter f blks], using the split function [f],
	 recursively splits the blocks in the list [blks]
	 into a list of buckets. Such splitting is performed
	 as much as possible.
       *)
      let rec split_iter f = function
	| [] -> []
	| e::els ->
	    match f e with
	      | Some(bucket, continuation) ->
		  if (Block.states continuation) = []
		  then bucket :: (split_iter f els)
		  else bucket :: (split_iter f (continuation::els))
	      | _         -> (split_iter f els) in

      let stop = ref false in
	
	while not ( !stop ) do
	  begin

	    (*
	      [oldblocks] records the blocks of the previous iteration.
	     *)
	    let oldblocks = !blocks in

	    let buckets = split_iter (split oldblocks) oldblocks in
	      begin

		(* The buckets computed by splitting all the blocks
		  are coerced to real blocks. Such coercion is performed
		  by adding to buckets the new information obtained in
		  the current iteration.
		*)
		blocks := (List.map (Block.close_block (env oldblocks)) buckets);

		(* The termination condition is evaluated.
		   The termination is reached when the current list of blocks
		   [blocks] is isomorphic to the list of blocks of the previous
		   iteration.

		   Note that if each block is not broken, then [i]-th block of the current
		   approximation ([blocks]) exactly corresponds to the [i]-th block of
		   previous approximation ([oldblocks]). Therefore, the comparison between
		   [blocks] and [oldblocks] can be done position-wise.
		*)
		stop := 
                   (List.length !blocks) = (List.length oldblocks) &&
		   (List.for_all2
		      (fun x y -> (Block.compare x y) == 0)
		      !blocks
		      oldblocks)

	      end
	  end
	done ;
	!blocks
(*i*) 	  
  let reduce a =

    (* NOTE: state_n creates a new state set at each invocation.. we 
     * need a smart method to change this situation
     *)

    (* compute all blocks in the automaton *)
    let blocks = partitioning a in

    let start = (compose (env blocks) (h_n blocks)) (Automaton.start a) in

    (* compute the states of the minimal automaton *)
    let states = snd (List.split (states_n blocks)) in

    (* arrows are obtained from the rapresentant element of each block *)
    let arrows = List.flatten 
		   (List.map 
		      (fun b -> 
			 Bundle.to_arrow_list (h_n blocks b) (Block.norm b))
			 blocks) in

      Automaton.create start states arrows
(*i*) 
end
