(* We will call "maps" a list of pairs that are the element
 * of a graph of discrete finite function
 *)

open Printf

let convertName id =
  if String.contains id '"' 
  then String.sub id 1 ((String.length id) - 2)
  else id


(* printing a list of integers *)
let print_int_list f l = 
  begin
    (fprintf f "[");
    (List.iter (fun x -> (fprintf f "%d ; " x)) l);
    (fprintf f "]");
  end


(* printing a map expressed as a list of pairs *)
let print_map_list f l = 
  begin
    (fprintf f "[");
    (List.iter (fun (x,y) -> (fprintf f "(%d->%d) ; " x y)) l);
    (fprintf f "]");
  end


(* let my_printf f = flush f; Printf.fprintf f *)


(* range 'lo' 'hi' constructs the segment of integers from
 * 'lo' to 'hi'. For example:
 *
 *        range 5 7 = [ 5; 6 ]
 *
 * Note that hi is not included.
 *)
let rec range lo hi =
  if hi<lo then failwith "range: hi<lo"
  else if lo == hi
  then []
  else lo :: (range (lo+1) hi)


(* list_rem 'el' 'list' returns the list obtained from 'list'
 * by removing all the occurrences of items equal to 'el'. Example:
 *
 *        list_rem 3 [ 3; 2 ; 4; 3 ] = [ 2; 4 ] 
 *
 *)
let rec list_rem el = function 
  |  []    -> []
  | e::es ->
      if (compare e el) == 0
      then list_rem el es
      else e :: (list_rem el es)


(* unique 'list' returns the list obtained by removing multiple
 * instances of elements of 'list'. For instance:
 *
 *        unique [ 1; 1; 2; 4; 3; 2 ] = [ 1; 2; 4; 3 ]
 *
 *)
let rec unique = function
  | [] -> []
  | e :: es ->
      if (List.mem e es)
      then unique es
      else e :: (unique es)


(* list_diff 'l' 'm' returns the list obtained by subtracting
 * 'm' from 'l'. Example:
 *        list_diff [ 3; 2; 4 ] [ 4; 3 ] = [ 2 ]
 *
 *)
let rec list_diff l = function
  | []      -> l
  | e :: es -> list_diff (list_rem e l) es


(* selections 'list' returns the list of pairs where the i-th
 * pair is (x_i, [x_1;..;x_i-1;x_i+1;..;x_n]).
 * For instance:
 *
 *        selections [ 1 ; 2 ; 3 ] = [(1, [2; 3]); (2, [1; 3]); (3, [1; 2])] 
 *
 * This function is used below to generate the permutations of
 * the elements in a list.
 * Those permutations should be thought of as permutations of elements
 * in a set. Therefore, we assume that
 *
 *        unique 'list' = 'list'
 *
 * i.e. each element of 'list' occurs only once in 'list'.
 *)
let rec selections = function
  | [] -> []
  | x::xs -> (x, xs) :: List.map (fun (x1,xs1)->(x1,(x::xs1))) (selections xs)


(* permutations_from_list 'list' exploits selections in orger to genereate
 * the list of permutations of elements in 'list'
 *
 *        permutations 3 = [ [1;2;3]; [1;3;2]; [2;1;3]; [2;3;1]; [3;1;2]; [3;2;1] ]
 *
 *)
let rec permutations_from_list = function
  | [] -> [ [] ]
  | xs -> List.flatten 
      (List.map 
	 (fun (x,xs) -> List.map (fun xs -> x::xs) 
	    (permutations_from_list xs))
	 (selections xs))

(* permutation 'n' generates the list containing all the permutaions
 * of the first 'n' integers.
 *)
let permutations_real n = 
  permutations_from_list (range 1 (n+1))

let __perms = Array.create 10 []

let permutations n =
  let result = Array.get __perms n in
    if result != []
    then begin (* Printf.printf "cached permutations for %d\n" n; *) result end
    else begin
      (* Printf.printf "computing permutations for %d\n" n; *)
      let perms_of_n = permutations_real n in
	Array.set __perms n perms_of_n; perms_of_n	
    end

(* to_fun 'list' takes a list of pairs and treat it as a function *)
let to_fun list =
  try
    (fun x -> List.assoc x list)
  with _ ->  failwith "to_fun: value not bounded"


(* dom 'map' returns the elements of the domain of 'map' *)
let dom map = fst (List.split map)


(* cod 'map' returnd the elements of the codomain of 'map' *)
let cod map = snd (List.split map)


(*  perm_to_fun 'rho ' 'x' returns the element in position 'i'
 * of 'rho'. For instance,
 *
 *        perm_to_fun [3; 2; 4; 1] 3 = 4
 *
 *)
let perm_to_fun rho = 
  fun i -> 
    if i == 0 
    then 0 
    else 
      begin
	try 
	  List.nth rho (i-1)
	with _ ->
	  failwith "perm_to_fun: value not bounded"
      end

(* index 'n' 'list' 'e' returns the position of the element 'e'
 * in 'list' summed to 'n'.
 *
 *        index 3 [ 3; 5; 6; 2] 2 = 7
 *
 * i.e. 3 + 4.
 * We assume that 'e' appears in 'list'.
 *)
let rec index n list e = 
  match list with
    | [] -> -1 (* failwith "Utils.index: element not found!" *)
    | x::xs -> 
	if e = x
	then n + 1
	else index (n+1) xs e


(* sub_list 'list' 'i' 'j' extracts the list containing the
 * elements of 'list' in the positions going from 'i' to 'j'.
 *)
let rec sub_list list i j = 
  try
    if (i == j)
    then [List.nth list i]
    else (List.nth list i) :: (sub_list list (i+1) j)
  with _ -> 
    failwith ("sub_list: first index greated than second one " ^
	      (string_of_int i) ^ ", " ^ (string_of_int j))


(* compose 'f' 'g' returns the function that maps an element
 * x in g(f(x))
 *)
let compose f g =  fun x -> g (f x)


(* compose_perm 'sigma' 'rho' compose a function and a permutation
 * of its codomain.
 *)
let rec compose_perm sigma rho =
   match sigma with
     | [] -> []
     | 0::is -> 0 :: (compose_perm is rho)
     | i::is -> (List.nth rho (i-1) ) :: (compose_perm is rho)


(* pos 'e' 'list' returns the first position in 'list' where 'e'
 * appears. If 'e' is not in 'list' then 0 is returned.
 *)
let rec pos e = function
    [] -> 0
  | el::els -> 
      if (e == el)
      then 1
      else 1 + (pos e els)


(* generate 'x' 'e' returns a list containing x elements each of one is e
 * e.g.
 *
 *        generate 4 'a' = [ 'a'; 'a'; 'a'; 'a' ]
 *
*)
let rec generate x e =
  if x < 0 
  then failwith "generate: x < 0" 
  else if x = 0 then []
  else e::(generate (x - 1) e)


(* invert_perm 'rho' returns the inverse of the permutation 'rho'.
 * For instance:
 *
 *        invert_perm [ 3; 2; 4; 1] = [4; 2; 1; 3]
 *
 *)
let invert_perm rho =
  List.map (index 0 rho) (range  1 (1 + (List.length rho)))


(* invert 'map' inverts a map. For example:
 *
 *        invert [ ( 3, 7); ( 2, 5); (6, 4) ] = [(7, 3); (5, 2); (4, 6)]
 *
 * Note that if 'map' is not bijective, then invert 'map' is not
 * a map!
 *)
let invert pairs =
  let (dom,cod) = List.split pairs in
    List.combine cod dom


(* permute 'list' 'rho' applies permutation 'rho' to a given list of elements.
 * Note that it is required that 'rho' and 'list' must have the same length.
 *)
let permute l rho =
  if (List.length l) != (List.length rho)
  then failwith "Utils.permute: lists have different length!"
  else List.map (fun i -> List.nth l (i-1)) rho







(*
 * The following functions are not used (right now).
 *)





(* check_perm 'list' takes a list of pairs and returns false iff
 * such pairs cannot be extended to a permutation
*)
let rec check_perm = function
    [] -> true
  | (u,v)::es ->
      let spl = (List.split es) in
      let dom = fun (x,y) -> not (x == u) || (y == v) in
      let cod = fun (x,y) -> not (y == v) || (x == u) in
      let domck = List.for_all dom es in
      let codck = List.for_all cod es in
	(check_perm es) && domck && codck
